
/*
Copyright 2011 Aurélien Bellet
aurelien.bellet@univ-st-etienne.fr

This file is part of GESL.

GESL is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GESL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GESL.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "GESL.h"

// load training/test sample
sample* load_data (char* data_dir, int nb_class) {
  int i,j,size;
  FILE *f;
  char line[1000];
  char path[1000];

  sample *data = (sample *) malloc(sizeof(sample));
  data->nb_class = nb_class;
  data->class = (sample_class **) malloc((data->nb_class)*sizeof(sample_class *));
  for (i=0; i<data->nb_class; i++) {
    data->class[i] = (sample_class *) malloc(sizeof(sample_class));
    sprintf(path,"%s/%d",data_dir,i);
    size = 0;
    f = fopen(path,"rt");
    if (f == NULL) {
      fprintf(stderr,"File %s does not exist.\n",path);
      exit(-1);
    }
    while (fgets(line,1000,f) != NULL)
      size++;
    fclose(f);
    data->class[i]->size = size;
    data->class[i]->t = (char **) malloc(data->class[i]->size*sizeof(char*));
    f = fopen(path,"rt");
    j=0;
    while (fgets(line,1000,f) != NULL) {
      data->class[i]->t[j] = (char *) malloc(1000*sizeof(char));
      sscanf(line,"%s\n",data->class[i]->t[j]);
      j++;
    }
    fclose(f);
  }

  return data;

}


// computes the Levenshtein edit distance
int levenshtein (char *x, char *y) {

  int i,j,lx,ly,ret;
  int **d;
  lx = strlen(x);
  ly = strlen(y);
  d = (int **) malloc((lx+1)*sizeof(int*));
  for (i=0; i<lx+1; i++)
    d[i] = (int *) malloc((ly+1)*sizeof(int));

  for (i=0; i<=lx; i++)
    d[i][0] = i;
  for (j=0; j<=ly; j++)
    d[0][j] = j;
  
  for (i=1; i<=lx; i++) {
    for (j=1; j<=ly; j++) {
      if (x[i-1] == y[j-1])
	d[i][j] = d[i-1][j-1];
      else
	d[i][j] = min(min(d[i-1][j]+1,d[i][j-1]+1),d[i-1][j-1]+1);
    }
  }
  
  ret = d[lx][ly];
  for (i=0; i<lx+1; i++)
    free(d[i]);
  free(d);
  return ret;

}


// computes the best edit script according to Levenshtein distance
// needed to build the optimization problem(s) and also to compute the learned similarity
int** script_lev (char *x, char *y, int alph_size, char ref_char) {

  int i,j,lx,ly;
  double **d;
  int **script;
  int diag,left,up;
  lx = strlen(x);
  ly = strlen(y);
  script = (int **) calloc((alph_size+1),sizeof(int*));
  for (i=0; i<=alph_size; i++)
    script[i] = (int *) calloc((alph_size+1),sizeof(int));
  d = (double **) malloc((lx+1)*sizeof(double*));
  for (i=0; i<lx+1; i++)
    d[i] = (double *) malloc((ly+1)*sizeof(double));

  for (i=0; i<=lx; i++)
    d[i][0] = i;
  for (j=0; j<=ly; j++)
    d[0][j] = j;

  for (i=1; i<=lx; i++) {
    for (j=1; j<=ly; j++) {
      if (x[i-1] == y[j-1])
	d[i][j] = d[i-1][j-1];
      else
	d[i][j] = min(min(d[i-1][j]+1,d[i][j-1]+1),d[i-1][j-1]+1);
    }
  }

  i = lx;
  j = ly;
  while (i != 0 && j != 0) {
    diag = d[i-1][j-1];
    left = d[i-1][j];
    up = d[i][j-1];
    if (diag <= left) {
      if (diag <= up) {
	i--;
	j--;
	if (x[i] != y[j]) {
	  script[x[i]-ref_char+1][y[j]-ref_char+1]++;
	}
      } else {
	j--;
	script[0][y[j]-ref_char+1]++;
      }
    } else if (left < up) {
      i--;
      script[x[i]-ref_char+1][0]++;
    } else {
      j--;
      script[0][y[j]-ref_char+1]++;
    }
  }

  while (i != 0) {
    i--;
    script[x[i]-ref_char+1][0]++;
  }

  while (j != 0) {
    j--;
    script[0][y[j]-ref_char+1]++;
  }

  for (i=0; i<lx+1; i++)
    free(d[i]);
  free(d);

  return script;

}


// save edit cost matrix for binary problem (class1,class2) to file output_dir/matrix_class1-class2
void save_matrix (char *output_dir, int class1, int class2, int alph_size, double **c) {
  int i,j;
  FILE *f;
  char path[1000];

  sprintf(path,"%s/matrix_%d-%d",output_dir,class1,class2);
  f = fopen(path,"wt");
  if (f == NULL) {
    fprintf(stderr,"Can't open file %s.\n",path);
    exit(-1);
  }

  for (i=0; i<=alph_size; i++) {
    for (j=0; j<=alph_size; j++) {
      fprintf(f,"%E ",c[i][j]);
    }
    fprintf(f,"\n");
  }

  fclose(f);

}


// compare function for pointers to ints
int cmp_intp(const void *a, const void *b) { 
  const int **ia = (const int **)a; 
  const int **ib = (const int **)b;
  return **ia  - **ib; 
}


// compare function that shuffles an array
int cmp_rand(const void *a, const void *b) { 
  int r = rand()%100;
  if (r >= 50)
    return 1;
  else
    return -1;
}


// checks if a given file exists
int file_exists (const char *filename) {
  FILE *file;
  if (file = fopen(filename, "r")) {
    fclose(file);
    return 1;
  }
  return 0;
}


// build training pairs with the Levenshtein distance
// and save the pairing in a file to avoid recomputing it each time
int**** build_levpairing (sample *data, double alpha, char *data_dir) {

  int i,j,k,p,q,z,sizep,sizeq;
  int nb_class = data->nb_class;
  int ****pairing;
  int *sim_vec;
  int **sim_vec_p;
  char path[1000];
  FILE *f;

  srand(1234);

  fprintf(stderr, "Pairing training examples (Levenshtein distance)... ");

  pairing = (int ****) calloc(data->nb_class,sizeof(int ***));

  sprintf(path,"%s/levpairing_done",data_dir);
  if (!file_exists(path)) {
    sprintf(path,"%s/levpairing.dat",data_dir);
    f = fopen(path,"wt");
    if (f == NULL) {
      fprintf(stderr,"Can't open file %s.\n",path);
      exit(-1);
    }

    for (p=0; p<nb_class; p++) {
      pairing[p] = (int ***) calloc(nb_class,sizeof(int **));
      for (q=0; q<nb_class; q++) {
	sizep = data->class[p]->size;
	sizeq = data->class[q]->size;
	pairing[p][q] = (int **) calloc(sizep,sizeof(int*));
	for (i=0; i<sizep; i++) {
	  pairing[p][q][i] = (int *) calloc(sizeq,sizeof(int));
	  sim_vec = (int *) malloc(sizeq*sizeof(int));
          sim_vec_p = (int **) malloc(sizeq*sizeof(int*));
	  for (j=0; j<sizeq; j++) {
	    sim_vec[j] = levenshtein(data->class[p]->t[i],data->class[q]->t[j]);
	    sim_vec_p[j] = &sim_vec[j];
	  }
	  qsort(sim_vec_p,sizeq,sizeof(sim_vec_p[0]),cmp_rand);
	  qsort(sim_vec_p,sizeq,sizeof(sim_vec_p[0]),cmp_intp);
	  for (j=0; j<sizeq; j++)
	    fprintf(f,"%d ",sim_vec_p[j]-sim_vec);
	  if (p == q)
	    for (j=0; j<(int)ceil(alpha*sizeq); j++)
	      if ((int)ceil(alpha*sizeq) == sizeq)
		pairing[p][q][i][sim_vec_p[j]-sim_vec] = 1;
	      else
		pairing[p][q][i][sim_vec_p[j+1]-sim_vec] = 1;
	  else
	    for (j=0; j<(int)ceil(alpha*sizeq); j++)
	      pairing[p][q][i][sim_vec_p[sizeq-j-1]-sim_vec] = 1;
          free(sim_vec);
          free(sim_vec_p);
	}
      }
    }
    fclose(f);
    sprintf(path,"%s/levpairing_done",data_dir);
    f = fopen(path,"wt");
    fprintf(f,"1\n");
    fclose(f);
  } else {
    sprintf(path,"%s/levpairing.dat",data_dir);
    f = fopen(path,"r");
    for (p=0; p<nb_class; p++) {
      pairing[p] = (int ***) calloc(nb_class,sizeof(int **));
      for (q=0; q<nb_class; q++) {
	sizep = data->class[p]->size;
	sizeq = data->class[q]->size;
	pairing[p][q] = (int **) calloc(sizep,sizeof(int*));
	for (i=0; i<sizep; i++) {
	  pairing[p][q][i] = (int *) calloc(sizeq,sizeof(int));
	  sim_vec = (int *) malloc(sizeq*sizeof(int));
	  z=0;
	  while (z < sizeq) {
	    fscanf(f,"%d ",&sim_vec[z]);
	    z++;
	  }
	  if (p == q)
	    for (j=0; j<(int)ceil(alpha*sizeq); j++)
	      if ((int)ceil(alpha*sizeq) == sizeq)
		pairing[p][q][i][sim_vec[j]] = 1;
	      else
		pairing[p][q][i][sim_vec[j+1]] = 1;
	  else
	    for (j=0; j<(int)ceil(alpha*sizeq); j++)
	      pairing[p][q][i][sim_vec[sizeq-j-1]] = 1;
          free(sim_vec);
	}
      }
    }
    fclose(f);
  }

  fprintf(stderr, "Done\n");

  return pairing;
}


// build the training pairs randomly
int**** build_randpairing (sample *data, double alpha) {

  int i,j,k,p,q,sizep,sizeq;
  int nb_class = data->nb_class;
  int ****pairing;
  int *vec;
  char path[1000];

  // fixed seed for reproducibility matters
  // can be changed to srand(time(NULL)) for different pairing at each call
  srand(1234);

  fprintf(stderr, "Pairing training examples (random)... ");

  pairing = (int ****) calloc(nb_class,sizeof(int ***));

  for (p=0; p<nb_class; p++) {
    pairing[p] = (int ***) calloc(nb_class,sizeof(int **));
    for (q=0; q<nb_class; q++) {
      sizep = data->class[p]->size;
      sizeq = data->class[q]->size;
      vec = (int *) malloc(sizeq*sizeof(int));
      for (i=0; i<sizeq; i++)
        vec[i] = i;
      pairing[p][q] = (int **) calloc(sizep,sizeof(int*));
      for (i=0; i<sizep; i++) {
        pairing[p][q][i] = (int *) calloc(sizeq,sizeof(int));
        qsort(vec,sizeq,sizeof(int),cmp_rand);
	if (p==q)
	  for (j=0; j<(int)ceil(alpha*sizeq); j++)
	    pairing[p][q][i][vec[j]]=1;
	else
	  for (j=0; j<(int)ceil(alpha*sizeq); j++)
	    pairing[p][q][i][vec[j]]=1;
      }
      free(vec);
    }
  }

  fprintf(stderr, "Done\n");

  return pairing;
}


// build the convex program(s) in MOSEK OPF format
void build_mosek (sample *data, int ****pairing, char* data_dir, double beta, double neta, int alph_size, char ref_char, double alpha, int norm_type) {
  
  int i,j,k,m,p,q,sizep,sizeq,total_pairs,nb_var,nb_cons;
  FILE *f;
  char path[1000];
  int **script;

  fprintf(stderr, "Building problem(s)... ");

  for (p=0; p<(data->nb_class)-1; p++) {
    for (q=p+1; q<data->nb_class; q++) {
      fprintf(stderr,"(%d,%d) ",p,q);

      sizep = data->class[p]->size;
      sizeq = data->class[q]->size;
      
      sprintf(path,"%s/%d-%d-%E-%E-%E.opf",data_dir,p,q,beta,neta,alpha);
      f = fopen(path,"wt");
      if (f == NULL) {
	fprintf(stderr,"Can't open file %s.\n",path);
	exit(-1);
      }

      total_pairs = ((int)ceil(alpha*sizep)+(int)ceil(alpha*sizeq))*(sizep+sizeq);

      // hints section
      nb_var = (alph_size+1)*(alph_size+1) + total_pairs + 3;
      nb_cons = total_pairs + ((alph_size+1)*(alph_size+1)-(alph_size+1))/2 + 1;
      fprintf(f,"[hints]\n  [hint NUMVAR] %d [/hint]\n  [hint NUMCON] %d [/hint]\n[/hints]\n\n",nb_var,nb_cons);

      // objective function
      fprintf(f,"[objective min 'f']\n");
      fprintf(f,"  %E ( ",beta*total_pairs);
      for (i=0; i<=alph_size; i++)
	for (j=0; j<=alph_size; j++)
	  if (i != alph_size || j != alph_size)
	    if (norm_type == 1)
	      fprintf(f,"c_%d_%d + ",i,j);
	    else
	      fprintf(f,"c_%d_%d^2 + ",i,j);
	  else
	    if (norm_type == 1)
	      fprintf(f,"c_%d_%d ) + ",i,j);
	    else
	      fprintf(f,"c_%d_%d^2 ) + ",i,j);
      for (i=0; i<sizep+sizeq; i++) {
	for (j=0; j<sizep+sizeq; j++) {
          if ((i < sizep && j<sizep && pairing[p][p][i][j] > 0) || (i >= sizep && j >= sizep && pairing[q][q][i-sizep][j-sizep] > 0))
            fprintf(f,"xi_%d_%d + ",i,j);
          else if ((i < sizep && j>=sizep && pairing[p][q][i][j-sizep] > 0) || (i >= sizep && j < sizep && pairing[q][p][i-sizep][j] > 0))
            fprintf(f,"xi_%d_%d + ",i,j);
	}
      }
      fprintf(f,"z\n");
      fprintf(f,"[/objective]\n\n");


      fprintf(f,"[constraints]\n");

      // constraints B1 - d(xi,xj) <= xi[i][j] for li != lj
      for (i=0; i<sizep; i++) {
	for (j=sizep; j<sizep+sizeq; j++) {
	  if (pairing[p][q][i][j-sizep] > 0) {
	    fprintf(f,"  [con] B1 - ");
	    script = script_lev(data->class[p]->t[i],data->class[q]->t[j-sizep],alph_size,ref_char);
	    for (k=0; k<=alph_size; k++)
	      for (m=0; m<=alph_size; m++)
	        if (script[k][m] != 0)
		  fprintf(f,"%d c_%d_%d - ",script[k][m],k,m);
	    fprintf(f,"xi_%d_%d <= 0 [/con]\n",i,j);
	  }
	}
      }
      for (i=sizep; i<sizep+sizeq; i++) {
	for (j=0; j<sizep; j++) {
	  if (pairing[q][p][i-sizep][j] > 0) {
	    fprintf(f,"  [con] B1 - ");
	    script = script_lev(data->class[q]->t[i-sizep],data->class[p]->t[j],alph_size,ref_char);
	    for (k=0; k<=alph_size; k++)
	      for (m=0; m<=alph_size; m++)
		if (script[k][m] != 0)
		fprintf(f,"%d c_%d_%d - ",script[k][m],k,m);
	    fprintf(f,"xi_%d_%d <= 0 [/con]\n",i,j);
	  }
	}
      }

      // constraints d(xi,xj) - B2 <= xi[i][j] for li == lj
      for (i=0; i<sizep; i++) {
	for (j=0; j<sizep; j++) {
	  if (pairing[p][p][i][j] > 0) {
	    script = script_lev(data->class[p]->t[i],data->class[p]->t[j],alph_size,ref_char);
	    fprintf(f,"  [con] ");
	    for (k=0; k<=alph_size; k++)
	      for (m=0; m<=alph_size; m++)
		if (script[k][m] != 0)
		  fprintf(f,"%d c_%d_%d + ",script[k][m],k,m);
	    fprintf(f,"z - B2 - xi_%d_%d <= 0 [/con]\n",i,j);
	  }
	}
      }
      for (i=sizep; i<sizep+sizeq; i++) {
	for (j=sizep; j<sizep+sizeq; j++) {
	  if (pairing[q][q][i-sizep][j-sizep] > 0) {
	    script = script_lev(data->class[q]->t[i-sizep],data->class[q]->t[j-sizep],alph_size,ref_char);
	    fprintf(f,"  [con] ");
	    for (k=0; k<=alph_size; k++)
	      for (m=0; m<=alph_size; m++)
		if (script[k][m] != 0)
		  fprintf(f,"%d c_%d_%d + ",script[k][m],k,m);
	    fprintf(f,"z - B2 - xi_%d_%d <= 0 [/con]\n",i,j);
	  }
	}
      }
      
      // constraint B1 - B2 = neta
      fprintf(f,"  [con] B1 - B2 = %E [/con]\n",neta);
      
     
      // constrain the cost matrix to be symmetric
      // may be removed if needed
      for (i=0; i<=alph_size; i++)
	for (j=i+1; j<=alph_size; j++)
	  fprintf(f,"  [con] c_%d_%d - c_%d_%d = 0 [/con]\n",i,j,j,i);

      fprintf(f,"[/constraints]\n\n");

      // variable bounds
      fprintf(f,"[bounds]\n  [b] 0 <= * [/b]\n  [b] z = 0 [/b]\n  [b] B1 >= 0.301 [/b]\n  [b] 0 <= B2 <= 0.301 [/b]\n[/bounds]\n");
     
      fclose(f);

    }
  }

  for (i=0; i<=alph_size; i++)
    free(script[i]);
  free(script);

  fprintf(stderr, "Done\n");
  
}


// solve the convex program(s) using MOSEK and save the resulting edit matrix(ces)
void solve_mosek (sample *data, char* data_dir, char* output_dir, int alph_size, double beta, double neta, double alpha) {
  int i,j,k,p,q,sizep,sizeq,total_pairs,nb_var,nb_cons;
  char path[1000],errname[1000],errdesc[1000];
  double **c,*var;
  MSKtask_t task;
  MSKenv_t env;
  MSKrescodee res;
  
  fprintf(stderr, "Solving problem(s) and saving matrix(ces) to %s ... ",output_dir);
  
  c = (double **) malloc((alph_size+1)*sizeof(double *));
  for (i=0; i<alph_size+1; i++)
    c[i] = (double *) malloc((alph_size+1)*sizeof(double));
  
  for (p=0; p<(data->nb_class)-1; p++) {
    for (q=p+1; q<data->nb_class; q++) {
      fprintf(stderr,"(%d,%d) ",p,q);
      
      sprintf(path,"%s/%d-%d-%E-%E-%E.opf",data_dir,p,q,beta,neta,alpha);

      sizep = data->class[p]->size;
      sizeq = data->class[q]->size;

      total_pairs = ((int)ceil(alpha*sizep)+(int)ceil(alpha*sizeq))*(sizep+sizeq);

      nb_var = (alph_size+1)*(alph_size+1) + total_pairs + 3;
      nb_cons = total_pairs + ((alph_size+1)*(alph_size+1)-(alph_size+1))/2 + 1;

      task = NULL;
      env = NULL;
      res = MSK_RES_OK;
      res = MSK_makeenv(&env,NULL,NULL,NULL,NULL);
      if (res==MSK_RES_OK)
	res = MSK_initenv(env);
      if (res==MSK_RES_OK)
	res = MSK_maketask(env,nb_cons,nb_var,&task);
      if (res==MSK_RES_OK)
	res = MSK_readdata(task,path);
      if (res==MSK_RES_OK)
	res = MSK_optimize(task);

      if (res != MSK_RES_OK) {
        // in case of a MOSEK error, print error code and description
        printf("\nMOSEK: An error occurred while optimizing.\n");
        MSK_getcodedesc(res,errname,errdesc);
        printf("Error %s - '%s'\n",errname,errdesc);
        exit(-1);
      } 

      /*
      // print final objective value
      double primalobj;
      MSK_getprimalobj (task,MSK_SOL_ITR,&primalobj);
      fprintf(stderr,"Objective function: %f\n",primalobj);
      */
      
      var = (double *) malloc((alph_size+1)*(alph_size+1)*sizeof(double));
      res = MSK_getsolutionslice(task,MSK_SOL_ITR,MSK_SOL_ITEM_XX,0,(alph_size+1)*(alph_size+1),var);

      MSK_deletetask(&task);
      MSK_deleteenv(&env);

      k=0;
      for (i=0; i<=alph_size; i++) {
	for (j=0; j<=alph_size; j++) {
	  if (var[k] == 0 || var[k] < 0)
	    c[i][j] = 0;
	  else
	    c[i][j] = var[k];
	  k++;
	}
      }

      free(var);

      // save the resulting matrix(ces)
      save_matrix(output_dir,p,q,alph_size,c);

      // delete the MOSEK OPF file
      sprintf(path,"rm -f %s/%d-%d-%E-%E-%E.*",data_dir,p,q,beta,neta,alpha);
      system(path);

    }
  }

  for (i=0; i<alph_size+1; i++)
    free(c[i]);
  free(c);
  
  fprintf(stderr, "Done\n");

}
