
/*
Copyright 2011 Aurélien Bellet
aurelien.bellet@univ-st-etienne.fr

This file is part of GESL.

GESL is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GESL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GESL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef HEADER_GESL
#define HEADER_GESL

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <mosek.h>

typedef struct sample_class {
  int size;
  char **t;
} sample_class;

typedef struct sample {
  int nb_class;
  sample_class **class;
} sample;

#define max(x1,x2) ((x1) > (x2) ? (x1):(x2))
#define min(x1,x2) ((x1) < (x2) ? (x1):(x2))
#define thresh(x) (((x) > 1) ? 1 : ((x) < -1) ? -1:(x))

// load training/test sample
sample* load_data (char* data_dir, int nb_class);

// computes the Levenshtein edit distance
int levenshtein (char *x, char *y);

// computes the best edit script according to Levenshtein distance
// needed to build the optimization problem(s) and also to compute the learned similarity
int** script_lev (char *x, char *y, int alph_size, char ref_char);

// save edit cost matrix for binary problem (class1,class2) to file output_dir/matrix_class1-class2
void save_matrix (char *output_dir, int class1, int class2, int alph_size, double **c);

// build training pairs with the Levenshtein distance
// and save the pairing in a file to avoid recomputing it each time
int**** build_levpairing (sample *data, double alpha, char *data_dir);

// build the training pairs randomly
int**** build_randpairing (sample *data, double alpha);

// build the convex program(s) in MOSEK OPF format
void build_mosek (sample *data, int ****pairing, char* data_dir, double beta, double neta, int alph_size, char ref_char, double alpha, int norm_type);

// solve the convex program(s) using MOSEK and save the resulting edit matrix(ces)
void solve_mosek (sample *data, char* data_dir, char* output_dir, int alph_size, double beta, double neta, double alpha);

#endif
