
/*
Copyright 2011 Aurélien Bellet
aurelien.bellet@univ-st-etienne.fr

This file is part of GESL.

GESL is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

GESL is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GESL.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef HEADER_GESL_CLASSIFY
#define HEADER_GESL_CLASSIFY

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "GESL.h"

// build Levenshtein similarity matrix(ces) from training data
double*** build_lev_mat (sample *data);

// return the (mean,std) values for each matrix
// used to normalize the Levenshtein distance to [-1,1]
double** normalize_mat (double*** sim_mat, sample *data);

// load edit matrix(ces) C learned with GESL
double*** load_matrices (char *matrix_dir, int alph_size, int nb_class);

// compute the similarity with respect to matrix c learned with GESL
// it first computes the Levenshtein script using script_lev() from "GESL.h"
// and applies custom costs c to it
double K_G (char *x, char *y, int alph_size, char ref_char, double **c);

// build GESL similarity matrix(ces) from training data
double*** build_GESL_mat (sample *data, int alph_size, char ref_char, double ***C);

// build linear program(s) in MOSEK LP format
void build_lp_mosek (sample *train, double ***sim_mat, double **norm, char* data_dir, int sim, double lambda);

// learn the classifier(s) alpha by solving the LP problems with MOSEK
// saves the classifier(s) to data_dir
double** solve_lp_mosek (char* data_dir, int nb_class, sample *train, int sim, double lambda);

// print the size of the learned classifier(s) alpha
void classifier_size (sample *train, double **alpha);

// classify test data according to learned classifier(s) alpha
// print confusion matrix and accuracy
void classify_test (sample *train, sample *test, int nb_class, double **norm, double **alpha, double ***C, int alph_size, char ref_char, int sim);


#endif
